from __future__ import annotations

from pathlib import Path
from typing import Optional

import typer

from svc_infra.db.nosql.scaffold import (
    scaffold_core,
    scaffold_documents_core,
    scaffold_resources_core,
    scaffold_schemas_core,
)


def cmd_scaffold(
    entity_name: str = typer.Option(
        "Item", help="Entity class name (e.g., User, Member, Product)."
    ),
    documents_dir: Path = typer.Option(..., help="Directory for Mongo document models."),
    schemas_dir: Path = typer.Option(..., help="Directory for Pydantic CRUD schemas."),
    overwrite: bool = typer.Option(False, help="Overwrite existing files."),
    same_dir: bool = typer.Option(
        False,
        "--same-dir/--no-same-dir",
        help="Put documents & schemas into the same directory.",
    ),
    documents_filename: Optional[str] = typer.Option(
        None, help="Custom filename for documents (separate-dir mode)."
    ),
    schemas_filename: Optional[str] = typer.Option(
        None, help="Custom filename for schemas (separate-dir mode)."
    ),
):
    """
    Scaffold starter Mongo document + CRUD schemas:
      • documents/<file>.py
      • schemas/<file>.py
    """
    res = scaffold_core(
        documents_dir=documents_dir,
        schemas_dir=schemas_dir,
        entity_name=entity_name,
        overwrite=overwrite,
        same_dir=same_dir,
        documents_filename=documents_filename,
        schemas_filename=schemas_filename,
    )
    typer.echo(res)


def cmd_scaffold_documents(
    dest_dir: Path = typer.Option(..., "--dest-dir", resolve_path=True),
    entity_name: str = typer.Option("Item", "--entity-name"),
    overwrite: bool = typer.Option(False, "--overwrite/--no-overwrite"),
    documents_filename: Optional[str] = typer.Option(
        None,
        "--documents-filename",
        help="Filename to write (e.g. product_doc.py). Defaults to <snake(entity)>.py",
    ),
):
    """Scaffold only the Mongo document model (Pydantic)."""
    res = scaffold_documents_core(
        dest_dir=dest_dir,
        entity_name=entity_name,
        overwrite=overwrite,
        documents_filename=documents_filename,
    )
    typer.echo(res)


def cmd_scaffold_schemas(
    dest_dir: Path = typer.Option(..., "--dest-dir", resolve_path=True),
    entity_name: str = typer.Option("Item", "--entity-name"),
    overwrite: bool = typer.Option(False, "--overwrite/--no-overwrite"),
    schemas_filename: Optional[str] = typer.Option(
        None,
        "--schemas-filename",
        help="Filename to write (e.g. product_schemas.py). Defaults to <snake(entity)>.py",
    ),
):
    """Scaffold only the CRUD schemas (Pydantic)."""
    res = scaffold_schemas_core(
        dest_dir=dest_dir,
        entity_name=entity_name,
        overwrite=overwrite,
        schemas_filename=schemas_filename,
    )
    typer.echo(res)


def cmd_scaffold_resources(
    dest_dir: Path = typer.Option(..., "--dest-dir", resolve_path=True),
    entity_name: str = typer.Option(
        "Item",
        "--entity-name",
        help="Used only to prefill example placeholders.",
    ),
    filename: Optional[str] = typer.Option(
        None,
        "--filename",
        help='Output filename (default: "resources.py")',
    ),
    overwrite: bool = typer.Option(False, "--overwrite/--no-overwrite"),
):
    """
    Scaffold a starter resources.py with an empty RESOURCES list.

    NOTE: Indexes are now declared directly on each NoSqlResource via its `indexes`
    attribute (e.g., `[IndexModel([...]), ...]`). There is no separate index-builders hook.
    """
    res = scaffold_resources_core(
        dest_dir=dest_dir,
        entity_name=entity_name,
        filename=filename,
        overwrite=overwrite,
    )
    typer.echo(res)


def register(app: typer.Typer) -> None:
    """
    Register Mongo scaffold commands on the given Typer app.
    Commands:
      • mongo-scaffold
      • mongo-scaffold-documents
      • mongo-scaffold-schemas
      • mongo-scaffold-resources
    """
    app.command("mongo-scaffold")(cmd_scaffold)
    app.command("mongo-scaffold-documents")(cmd_scaffold_documents)
    app.command("mongo-scaffold-schemas")(cmd_scaffold_schemas)
    app.command("mongo-scaffold-resources")(cmd_scaffold_resources)



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Literal, Optional

from svc_infra.db.utils import normalize_dir, pascal, plural_snake, snake
from svc_infra.utils import ensure_init_py, render_template, write

# ---------------- helpers ----------------

_INIT_CONTENT_PAIRED = 'from . import models, schemas\n\n__all__ = ["models", "schemas"]\n'
_INIT_CONTENT_MINIMAL = "# package marker; add explicit exports here if desired\n"


def _ensure_init_py(dir_path: Path, overwrite: bool, paired: bool) -> Dict[str, Any]:
    """Create __init__.py; paired=True writes models/schemas re-exports, otherwise minimal."""
    content = _INIT_CONTENT_PAIRED if paired else _INIT_CONTENT_MINIMAL
    return ensure_init_py(dir_path, overwrite, paired, content)


# ---------------- unified public API ----------------
# kind: "entity" (generic) or "auth" (specialized templates)

Kind = Literal["entity", "auth"]


def scaffold_core(
    *,
    models_dir: Path | str,
    schemas_dir: Path | str,
    kind: Kind = "entity",
    entity_name: str = "Item",
    table_name: Optional[str] = None,
    include_tenant: bool = True,
    include_soft_delete: bool = False,
    overwrite: bool = False,
    same_dir: bool = False,
    models_filename: Optional[str] = None,  # <--- NEW
    schemas_filename: Optional[str] = None,  # <--- NEW
) -> Dict[str, Any]:
    """
    Create starter model + schema files.

    Filenames:
      - same_dir=True  -> models.py + schemas.py (paired).
      - same_dir=False -> defaults to <snake(entity)>.py in each dir, unless you pass
                          --models-filename / --schemas-filename.
    """
    models_dir = normalize_dir(models_dir)
    schemas_dir = normalize_dir(models_dir if same_dir else schemas_dir)

    # content per kind
    if kind == "auth":
        auth_ent = pascal(entity_name or "User")
        auth_tbl = table_name or plural_snake(auth_ent)

        models_txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.auth",
            name="models.py.tmpl",
            subs={"AuthEntity": auth_ent, "auth_table_name": auth_tbl},
        )
        schemas_txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.auth",
            name="schemas.py.tmpl",
            subs={"AuthEntity": auth_ent},
        )
    else:
        ent = pascal(entity_name)
        tbl = table_name or plural_snake(ent)

        tenant_model_field = (
            "    tenant_id: Mapped[Optional[str]] = mapped_column(String(64), index=True)\n"
            if include_tenant
            else ""
        )
        soft_delete_model_field = (
            "    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)\n"
            "    deleted_at: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))\n"
            if include_soft_delete
            else "    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)\n"
        )

        tenant_arg = ', tenant_field="tenant_id"' if include_tenant else ""
        tenant_default = '"tenant_id"' if include_tenant else "None"

        models_txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.entity",
            name="models.py.tmpl",
            subs={
                "Entity": ent,
                "table_name": tbl,
                "tenant_field": tenant_model_field,
                "soft_delete_field": soft_delete_model_field,
                "tenant_arg": tenant_arg,
                "tenant_default": tenant_default,
            },
        )

        tenant_schema_field = "    tenant_id: Optional[str] = None\n" if include_tenant else ""
        schemas_txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.entity",
            name="schemas.py.tmpl",
            subs={
                "Entity": ent,
                "tenant_field": tenant_schema_field,
                "tenant_field_create": tenant_schema_field,
                "tenant_field_update": tenant_schema_field,
            },
        )

    # filenames
    if same_dir:
        models_path = models_dir / "models.py"
        schemas_path = schemas_dir / "schemas.py"
    else:
        default_stub = snake(entity_name)
        models_name = models_filename or f"{default_stub}.py"
        schemas_name = schemas_filename or f"{default_stub}.py"
        models_path = models_dir / models_name
        schemas_path = schemas_dir / schemas_name

    # write
    models_res = write(models_path, models_txt, overwrite)
    schemas_res = write(schemas_path, schemas_txt, overwrite)

    # __init__ files
    init_results = []
    init_results.append(_ensure_init_py(models_dir, overwrite, paired=same_dir))
    if schemas_dir != models_dir:
        init_results.append(_ensure_init_py(schemas_dir, overwrite, paired=False))

    return {
        "status": "ok",
        "results": {
            "models": models_res,
            "schemas": schemas_res,
            "inits": init_results,
        },
    }


def scaffold_models_core(
    *,
    dest_dir: Path | str,
    kind: Kind = "entity",
    entity_name: str = "Item",
    table_name: Optional[str] = None,
    include_tenant: bool = True,
    include_soft_delete: bool = False,
    overwrite: bool = False,
    models_filename: Optional[str] = None,  # <--- NEW
) -> Dict[str, Any]:
    """Create only a model file (defaults to <snake(entity)>.py unless models_filename is provided)."""
    dest = normalize_dir(dest_dir)

    if kind == "auth":
        auth_ent = pascal(entity_name or "User")
        auth_tbl = table_name or plural_snake(auth_ent)
        txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.auth",
            name="models.py.tmpl",
            subs={"AuthEntity": auth_ent, "auth_table_name": auth_tbl},
        )
    else:
        ent = pascal(entity_name)
        tbl = table_name or plural_snake(ent)

        tenant_model_field = (
            "    tenant_id: Mapped[Optional[str]] = mapped_column(String(64), index=True)\n"
            if include_tenant
            else ""
        )
        soft_delete_model_field = (
            "    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)\n"
            "    deleted_at: Mapped[Optional[datetime]] = mapped_column(DateTime(timezone=True))\n"
            if include_soft_delete
            else "    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)\n"
        )
        tenant_arg = ', tenant_field="tenant_id"' if include_tenant else ""
        tenant_default = '"tenant_id"' if include_tenant else "None"
        txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.entity",
            name="models.py.tmpl",
            subs={
                "Entity": ent,
                "table_name": tbl,
                "tenant_field": tenant_model_field,
                "soft_delete_field": soft_delete_model_field,
                "tenant_arg": tenant_arg,
                "tenant_default": tenant_default,
            },
        )

    filename = models_filename or f"{snake(entity_name)}.py"
    res = write(dest / filename, txt, overwrite)
    _ensure_init_py(dest, overwrite, paired=False)
    return {"status": "ok", "result": res}


def scaffold_schemas_core(
    *,
    dest_dir: Path | str,
    kind: Kind = "entity",
    entity_name: str = "Item",
    include_tenant: bool = True,
    overwrite: bool = False,
    schemas_filename: Optional[str] = None,  # <--- NEW
) -> Dict[str, Any]:
    """Create only a schema file (defaults to <snake(entity)>.py unless schemas_filename is provided)."""
    dest = normalize_dir(dest_dir)

    if kind == "auth":
        auth_ent = pascal(entity_name or "User")
        txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.auth",
            name="schemas.py.tmpl",
            subs={"AuthEntity": auth_ent},
        )
    else:
        ent = pascal(entity_name)
        tenant_field = "    tenant_id: Optional[str] = None\n" if include_tenant else ""
        txt = render_template(
            tmpl_dir="svc_infra.db.sql.templates.models_schemas.entity",
            name="schemas.py.tmpl",
            subs={
                "Entity": ent,
                "tenant_field": tenant_field,
                "tenant_field_create": tenant_field,
                "tenant_field_update": tenant_field,
            },
        )

    filename = schemas_filename or f"{snake(entity_name)}.py"
    res = write(dest / filename, txt, overwrite)
    _ensure_init_py(dest, overwrite, paired=False)
    return {"status": "ok", "result": res}



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Iterable, Optional, Sequence, Type, Union

from pymongo import IndexModel


def _snake(name: str) -> str:
    import re

    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)
    return re.sub(r"[^a-zA-Z0-9_]+", "_", s2).lower().strip("_")


def _default_collection_for(model: type) -> str:
    base = _snake(getattr(model, "__name__", "item"))
    return base if base.endswith("s") else base + "s"


def get_collection_name(document_model: type) -> str:
    name = getattr(document_model, "__collection__", None)
    return (
        name.strip()
        if isinstance(name, str) and name.strip()
        else _default_collection_for(document_model)
    )


IndexAlias = dict[str, Any]  # dict alias normalized to IndexModel later


@dataclass
class NoSqlResource:
    """
    Mongo resource declaration used by API & CLI.
    Define indexes here (either IndexModel or simple alias dicts).
    """

    # API mounting
    collection: Optional[str] = None
    prefix: str = ""
    document_model: Optional[Type[Any]] = None

    # optional Pydantic schemas (auto-derived if omitted)
    read_schema: Optional[Type[Any]] = None
    create_schema: Optional[Type[Any]] = None
    update_schema: Optional[Type[Any]] = None

    # behavior
    search_fields: Optional[Sequence[str]] = None
    tags: Optional[list[str]] = None
    id_field: str = "_id"
    soft_delete: bool = False
    soft_delete_field: str = "deleted_at"
    soft_delete_flag_field: Optional[str] = None

    # custom wiring
    service_factory: Optional[Callable[[Any], Any]] = None

    # generated schema naming and exclusions
    read_name: Optional[str] = None
    create_name: Optional[str] = None
    update_name: Optional[str] = None
    create_exclude: tuple[str, ...] = ("_id",)
    read_exclude: tuple[str, ...] = ()
    update_exclude: tuple[str, ...] = ()

    # NEW: indexes defined per collection (normalized to IndexModel at prepare time)
    indexes: Optional[Iterable[Union[IndexModel, IndexAlias]]] = None

    def __post_init__(self):
        if not self.collection and self.document_model:
            self.collection = get_collection_name(self.document_model)

    def resolved_collection(self) -> str:
        if self.collection:
            return self.collection
        if self.document_model:
            return get_collection_name(self.document_model)
        raise ValueError(
            "No collection name resolved. Set `collection=` or define a document_model with __collection__."
        )



from __future__ import annotations

from typing import Any, Iterable, Optional, Sequence, Set

from sqlalchemy import Select, String, and_, func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute, class_mapper


class SqlRepository:
    """
    Very small async repository around a mapped SQLAlchemy model.
    """

    def __init__(
        self,
        *,
        model: type[Any],
        id_attr: str = "id",
        soft_delete: bool = False,
        soft_delete_field: str = "deleted_at",
        soft_delete_flag_field: str | None = None,
        immutable_fields: Optional[Set[str]] = None,
    ):
        self.model = model
        self.id_attr = id_attr
        self.soft_delete = soft_delete
        self.soft_delete_field = soft_delete_field
        self.soft_delete_flag_field = soft_delete_flag_field
        self.immutable_fields: Set[str] = set(
            immutable_fields or {"id", "created_at", "updated_at"}
        )

    def _model_columns(self) -> set[str]:
        return {c.key for c in class_mapper(self.model).columns}

    def _id_column(self) -> InstrumentedAttribute:
        return getattr(self.model, self.id_attr)

    def _base_select(self) -> Select:
        stmt = select(self.model)
        if self.soft_delete:
            # Filter out soft-deleted rows by timestamp and/or active flag
            if hasattr(self.model, self.soft_delete_field):
                stmt = stmt.where(getattr(self.model, self.soft_delete_field).is_(None))
            if self.soft_delete_flag_field and hasattr(self.model, self.soft_delete_flag_field):
                stmt = stmt.where(getattr(self.model, self.soft_delete_flag_field).is_(True))
        return stmt

    # basic ops

    async def list(
        self,
        session: AsyncSession,
        *,
        limit: int,
        offset: int,
        order_by: Optional[Sequence[Any]] = None,
    ) -> Sequence[Any]:
        stmt = self._base_select().limit(limit).offset(offset)
        if order_by:
            stmt = stmt.order_by(*order_by)
        rows = (await session.execute(stmt)).scalars().all()
        return rows

    async def count(self, session: AsyncSession) -> int:
        stmt = select(func.count()).select_from(self._base_select().subquery())
        return (await session.execute(stmt)).scalar_one()

    async def get(self, session: AsyncSession, id_value: Any) -> Any | None:
        # honors soft-delete if configured
        stmt = self._base_select().where(self._id_column() == id_value)
        return (await session.execute(stmt)).scalars().first()

    async def create(self, session: AsyncSession, data: dict[str, Any]) -> Any:
        valid = self._model_columns()
        filtered = {k: v for k, v in data.items() if k in valid}
        obj = self.model(**filtered)
        session.add(obj)
        await session.flush()
        return obj

    async def update(
        self, session: AsyncSession, id_value: Any, data: dict[str, Any]
    ) -> Any | None:
        obj = await self.get(session, id_value)
        if not obj:
            return None
        valid = self._model_columns()
        for k, v in data.items():
            if k in valid and k not in self.immutable_fields:
                setattr(obj, k, v)
        await session.flush()
        return obj

    async def delete(self, session: AsyncSession, id_value: Any) -> bool:
        obj = await session.get(self.model, id_value)
        if not obj:
            return False
        if self.soft_delete:
            # Prefer timestamp, also optionally set flag to False
            if hasattr(self.model, self.soft_delete_field):
                setattr(obj, self.soft_delete_field, func.now())
            if self.soft_delete_flag_field and hasattr(self.model, self.soft_delete_flag_field):
                setattr(obj, self.soft_delete_flag_field, False)
            await session.flush()
            return True
        await session.delete(obj)
        await session.flush()
        return True

    async def search(
        self,
        session: AsyncSession,
        *,
        q: str,
        fields: Sequence[str],
        limit: int,
        offset: int,
        order_by: Optional[Sequence[Any]] = None,
    ) -> Sequence[Any]:
        ilike = f"%{q}%"
        conditions = []
        for f in fields:
            col = getattr(self.model, f, None)
            if col is not None:
                try:
                    conditions.append(col.cast(String).ilike(ilike))
                except Exception:
                    # skip columns that cannot be used in ilike even with cast
                    continue
        stmt = self._base_select()
        if conditions:
            stmt = stmt.where(or_(*conditions))
        stmt = stmt.limit(limit).offset(offset)
        if order_by:
            stmt = stmt.order_by(*order_by)
        return (await session.execute(stmt)).scalars().all()

    async def count_filtered(self, session: AsyncSession, *, q: str, fields: Sequence[str]) -> int:
        ilike = f"%{q}%"
        conditions = []
        for f in fields:
            col = getattr(self.model, f, None)
            if col is not None:
                try:
                    conditions.append(col.cast(String).ilike(ilike))
                except Exception:
                    continue
        stmt = self._base_select()
        if conditions:
            stmt = stmt.where(or_(*conditions))
        # SELECT COUNT(*) FROM (<stmt>) as t
        return (
            await session.execute(select(func.count()).select_from(stmt.subquery()))
        ).scalar_one()

    async def exists(self, session: AsyncSession, *, where: Iterable[Any]) -> bool:
        stmt = self._base_select().where(and_(*where)).limit(1)
        return (await session.execute(stmt)).first() is not None



from __future__ import annotations

import hashlib
import hmac
import os
import uuid
from datetime import datetime, timezone

from sqlalchemy import JSON, Boolean, DateTime, ForeignKey, Index, String, UniqueConstraint, text
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import Mapped, mapped_column, relationship

from svc_infra.db.sql.base import ModelBase
from svc_infra.db.sql.types import GUID

_APIKEY_HMAC_SECRET = os.getenv("APIKEY_HASH_SECRET") or "change-me-low-entropy-dev"


def _hmac_sha256(s: str) -> str:
    return hmac.new(_APIKEY_HMAC_SECRET.encode(), s.encode(), hashlib.sha256).hexdigest()


def _now() -> datetime:
    return datetime.now(timezone.utc)


class ApiKey(ModelBase):
    __tablename__ = "api_keys"

    id: Mapped[uuid.UUID] = mapped_column(GUID(), primary_key=True, default=uuid.uuid4)

    # Optional owner (bind key to a user). If null => service key.
    user_id: Mapped[uuid.UUID | None] = mapped_column(
        GUID(), ForeignKey("users.id", ondelete="SET NULL")
    )
    user = relationship("User", lazy="selectin")

    name: Mapped[str] = mapped_column(String(128), nullable=False)
    # We store only a hash + a short prefix for quick lookup/UX
    key_prefix: Mapped[str] = mapped_column(String(12), index=True, nullable=False)
    key_hash: Mapped[str] = mapped_column(String(64), nullable=False)  # hex sha256

    scopes: Mapped[list[str]] = mapped_column(MutableList.as_mutable(JSON), default=list)
    active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    expires_at: Mapped[datetime | None] = mapped_column(DateTime(timezone=True))
    last_used_at: Mapped[datetime | None] = mapped_column(DateTime(timezone=True))
    meta: Mapped[dict] = mapped_column(MutableDict.as_mutable(JSON), default=dict)

    created_at = mapped_column(
        DateTime(timezone=True), server_default=text("CURRENT_TIMESTAMP"), nullable=False
    )
    updated_at = mapped_column(
        DateTime(timezone=True),
        server_default=text("CURRENT_TIMESTAMP"),
        onupdate=text("CURRENT_TIMESTAMP"),
        nullable=False,
    )

    __table_args__ = (
        UniqueConstraint("key_prefix", name="uq_apikey_prefix"),
        Index("ix_api_keys_user_id", "user_id"),
    )

    # Helpers
    @staticmethod
    def make_secret() -> tuple[str, str, str]:
        """
        Returns (plaintext, prefix, hash). The plaintext is shown ONCE to the caller.
        Format: ak_<prefix>_<random>
        """
        import base64
        import secrets

        prefix = secrets.token_urlsafe(6).replace("-", "").replace("_", "")[:8]
        rand = base64.urlsafe_b64encode(secrets.token_bytes(24)).decode().rstrip("=")
        plaintext = f"ak_{prefix}_{rand}"
        return plaintext, prefix, _hmac_sha256(plaintext)

    @staticmethod
    def hash(plaintext: str) -> str:
        return _hmac_sha256(plaintext)

    def mark_used(self):
        self.last_used_at = _now()



from __future__ import annotations

from typing import Any, Iterable, List, Optional, Tuple, Type

from sqlalchemy import Index, func

from svc_infra.db.utils import KeySpec
from svc_infra.db.utils import as_tuple as _as_tuple


def make_unique_sql_indexes(
    model: Type[Any],
    *,
    unique_cs: Iterable[KeySpec] = (),
    unique_ci: Iterable[KeySpec] = (),
    tenant_field: Optional[str] = None,
    name_prefix: str = "uq",
) -> List[Index]:
    """Return SQLAlchemy Index objects that enforce uniqueness.

    - unique_cs: case-sensitive unique specs
    - unique_ci: case-insensitive unique specs (lower(column))
    - tenant_field: if provided, create two partial unique indexes:
        * tenant IS NULL (global bucket)
        * tenant IS NOT NULL (scoped per-tenant)

    Declare right after your model class; Alembic or metadata.create_all will pick them up.
    """
    idxs: List[Index] = []

    def _col(name: str):
        return getattr(model, name)

    def _to_sa_cols(spec: Tuple[str, ...], *, ci: bool):
        cols = []
        for cname in spec:
            c = _col(cname)
            cols.append(func.lower(c) if ci else c)
        return tuple(cols)

    tenant_col = _col(tenant_field) if tenant_field else None

    def _name(ci: bool, spec: Tuple[str, ...], null_bucket: Optional[str] = None):
        parts = [name_prefix, model.__tablename__]
        if tenant_field:
            parts.append(tenant_field)
        if null_bucket:
            parts.append(null_bucket)
        parts.append("ci" if ci else "cs")
        parts.extend(spec)
        return "_".join(parts)

    for ci, spec_list in ((False, unique_cs), (True, unique_ci)):
        for spec in spec_list:
            spec_t = _as_tuple(spec)
            cols = _to_sa_cols(spec_t, ci=ci)

            if tenant_col is None:
                idxs.append(Index(_name(ci, spec_t), *cols, unique=True))
            else:
                idxs.append(
                    Index(
                        _name(ci, spec_t, "null"),
                        *cols,
                        unique=True,
                        postgresql_where=tenant_col.is_(None),
                    )
                )
                idxs.append(
                    Index(
                        _name(ci, spec_t, "notnull"),
                        tenant_col,
                        *cols,
                        unique=True,
                        postgresql_where=tenant_col.isnot(None),
                    )
                )
    return idxs



from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from pymongo import ASCENDING, DESCENDING, IndexModel
from pymongo.collation import Collation

# Devs can hand you either a ready IndexModel or a dict alias. We normalize to IndexModel.
# Dict alias shape:
# {
#   "keys": [("field", 1), ("other", -1)]  # or ["field", "-other"]
#   "unique": True/False,
#   "name": "idx_name",
#   "partialFilterExpression": {...},
#   "expireAfterSeconds": 3600,
#   "collation": {"locale":"en", "strength":2},
#   "sparse": True/False,
#   "background": True/False   # ignored by MongoDB 6+, harmless to pass
# }
Alias = Dict[str, Any]
KeySpec = Union[str, Tuple[str, int]]


def _normalize_keys(keys: Iterable[KeySpec]) -> List[Tuple[str, int]]:
    out: List[Tuple[str, int]] = []
    for k in keys:
        if isinstance(k, tuple):
            field, dir_val = k
            direction = ASCENDING if dir_val >= 0 else DESCENDING
            out.append((field, direction))
        else:
            # string form: "field" => ASC; "-field" => DESC
            name = k[1:] if k.startswith("-") else k
            direction = DESCENDING if k.startswith("-") else ASCENDING
            out.append((name, direction))
    return out


def _normalize_collation(c: Optional[Dict[str, Any]]) -> Optional[Collation]:
    if not c:
        return None
    # common short form e.g. {"locale":"en","strength":2}
    return Collation(**c)


def normalize_index(idx: Union[IndexModel, Alias]) -> IndexModel:
    if isinstance(idx, IndexModel):
        return idx
    keys = _normalize_keys(idx.get("keys", []))
    if not keys:
        raise ValueError("Index alias requires 'keys'.")
    kwargs: Dict[str, Any] = {}
    for k in (
        "name",
        "unique",
        "partialFilterExpression",
        "sparse",
        "expireAfterSeconds",
        "background",
    ):
        if k in idx:
            kwargs[k] = idx[k]
    collation = _normalize_collation(idx.get("collation"))
    if collation:
        kwargs["collation"] = collation
    return IndexModel(keys, **kwargs)


def normalize_indexes(indexes: Optional[Iterable[Union[IndexModel, Alias]]]) -> List[IndexModel]:
    if not indexes:
        return []
    return [normalize_index(i) for i in indexes]



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import List, Optional

from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import BaseModel, Field
from sqlalchemy import select

from svc_infra.api.fastapi import public_router
from svc_infra.api.fastapi.auth.security import current_principal
from svc_infra.api.fastapi.db.sql.session import SqlSessionDep
from svc_infra.db.sql.apikey import ApiKey


class ApiKeyCreateIn(BaseModel):
    name: str
    user_id: Optional[str] = None  # if omitted => bound to caller or service key if caller is None
    scopes: List[str] = Field(default_factory=list)
    ttl_hours: Optional[int] = 24 * 365  # default 1y


class ApiKeyOut(BaseModel):
    id: str
    name: str
    user_id: Optional[str]
    key: Optional[str] = None  # ONLY present at creation time
    key_prefix: str
    scopes: List[str]
    active: bool
    expires_at: Optional[datetime]
    last_used_at: Optional[datetime]


def apikey_router(prefix: str = "/auth/keys") -> APIRouter:
    r = public_router(prefix=prefix, tags=["auth:apikeys"])

    @r.post(
        "",
        response_model=ApiKeyOut,
        openapi_extra={
            "security": [
                {"OAuth2PasswordBearer": []},
                {"cookieAuth": []},
                {"APIKeyHeader": []},
            ]
        },
    )
    async def create_key(
        sess: SqlSessionDep,
        payload: ApiKeyCreateIn = Body(...),
        p=Depends(current_principal),
    ):
        # require superuser or owner policy as you prefer
        if not p.user or not getattr(p.user, "is_superuser", False):
            raise HTTPException(403, "forbidden")

        plaintext, prefix, hashed = ApiKey.make_secret()
        expires = (
            (datetime.now(timezone.utc) + timedelta(hours=payload.ttl_hours))
            if payload.ttl_hours
            else None
        )

        row = ApiKey(
            user_id=payload.user_id or (getattr(p.user, "id", None)),
            name=payload.name,
            key_prefix=prefix,
            key_hash=hashed,
            scopes=payload.scopes,
            active=True,
            expires_at=expires,
        )
        sess.add(row)
        await sess.flush()
        return ApiKeyOut(
            id=str(row.id),
            name=row.name,
            user_id=str(row.user_id) if row.user_id else None,
            key=plaintext,  # show ONCE
            key_prefix=row.key_prefix,
            scopes=row.scopes,
            active=row.active,
            expires_at=row.expires_at,
            last_used_at=row.last_used_at,
        )

    @r.get(
        "",
        response_model=list[ApiKeyOut],
        openapi_extra={
            "security": [
                {"OAuth2PasswordBearer": []},
                {"cookieAuth": []},
                {"APIKeyHeader": []},
            ]
        },
    )
    async def list_keys(sess: SqlSessionDep, p=Depends(current_principal)):
        if not p.user or not getattr(p.user, "is_superuser", False):
            raise HTTPException(403, "forbidden")
        rows = (await sess.execute(select(ApiKey))).scalars().all()
        return [
            ApiKeyOut(
                id=str(x.id),
                name=x.name,
                user_id=str(x.user_id) if x.user_id else None,
                key=None,
                key_prefix=x.key_prefix,
                scopes=x.scopes,
                active=x.active,
                expires_at=x.expires_at,
                last_used_at=x.last_used_at,
            )
            for x in rows
        ]

    @r.post(
        "/{key_id}/revoke",
        openapi_extra={
            "security": [
                {"OAuth2PasswordBearer": []},
                {"cookieAuth": []},
                {"APIKeyHeader": []},
            ]
        },
    )
    async def revoke_key(key_id: str, sess: SqlSessionDep, p=Depends(current_principal)):
        if not p.user or not getattr(p.user, "is_superuser", False):
            raise HTTPException(403, "forbidden")
        row = await sess.get(ApiKey, key_id)
        if not row:
            raise HTTPException(404, "not_found")
        row.active = False
        await sess.commit()
        return {"ok": True}

    return r



from svc_infra.api.fastapi.db.sql.add import add_sql_db, add_sql_health, add_sql_resources
from svc_infra.api.fastapi.db.sql.session import SqlSessionDep

__all__ = [
    "SqlSessionDep",
    "add_sql_health",
    "add_sql_db",
    "add_sql_resources",
]



from __future__ import annotations

import importlib
import os
from typing import Any, Optional, Sequence

import typer

# Reuse core logic (all validation/locking happens there)
from svc_infra.db.nosql.core import prepare_mongo, setup_and_prepare

# Client lifecycle for the async command
from svc_infra.db.nosql.mongo.client import close_mongo, init_mongo
from svc_infra.db.nosql.resource import NoSqlResource
from svc_infra.db.nosql.utils import prepare_process_env

# -------------------- helpers --------------------


def _apply_mongo_env(mongo_url: Optional[str], mongo_db: Optional[str]) -> None:
    """If provided, set MONGO_URL / MONGO_DB for the current process."""
    if mongo_url:
        os.environ["MONGO_URL"] = mongo_url
    if mongo_db:
        os.environ["MONGO_DB"] = mongo_db


def _load_obj(dotted: str) -> Any:
    """
    Load an object from a dotted path like:
      - 'pkg.mod:NAME'  (preferred)
      - 'pkg.mod.NAME'  (also accepted)
    """
    if ":" in dotted:
        mod_path, attr = dotted.split(":", 1)
    else:
        mod_path, _, attr = dotted.rpartition(".")
        if not mod_path:
            raise ValueError(f"Invalid dotted path: {dotted}")
    mod = importlib.import_module(mod_path)
    try:
        return getattr(mod, attr)
    except AttributeError as e:
        raise ValueError(f"Object {attr!r} not found in module {mod_path!r}") from e


def _normalize_resources(obj: Any) -> Sequence[NoSqlResource]:
    if obj is None:
        raise ValueError("No resources provided.")
    if isinstance(obj, NoSqlResource):
        return [obj]
    if isinstance(obj, (list, tuple)):
        return obj  # best-effort
    raise TypeError("resources must be a NoSqlResource or a sequence of them")


# -------------------- commands --------------------


def cmd_prepare(
    resources_path: str = typer.Option(
        ...,
        "--resources",
        help="Dotted path to NoSqlResource(s). e.g. 'app.db.mongo:RESOURCES'",
    ),
    mongo_url: Optional[str] = typer.Option(
        None, "--mongo-url", help="Overrides env MONGO_URL for this command."
    ),
    mongo_db: Optional[str] = typer.Option(
        None, "--mongo-db", help="Overrides env MONGO_DB for this command."
    ),
    service_id: Optional[str] = typer.Option(
        None,
        "--service-id",
        help="Stable ID for this service/app. Defaults to top-level module name.",
    ),
    allow_rebind: bool = typer.Option(
        False, "--allow-rebind", help="Permit moving to a different DB."
    ),
):
    """
    Ensure Mongo is reachable, collections exist, and indexes (from each NoSqlResource.indexes)
    are applied.

    This command is async (uses Motor). We set env overrides and bootstrap .env,
    open the client, then delegate validation/locking to core.prepare_mongo().
    """
    _apply_mongo_env(mongo_url, mongo_db)
    prepare_process_env(".")
    resources = _normalize_resources(_load_obj(resources_path))
    sid = service_id or _default_service_id_from_resources_path(resources_path)

    import asyncio

    async def _run():
        await init_mongo()
        try:
            result = await prepare_mongo(
                resources=resources,
                service_id=sid,
                allow_rebind=allow_rebind,
            )
            return {
                "ok": result.ok,
                "created_collections": result.created_collections,
                "created_indexes": result.created_indexes,
            }
        finally:
            await close_mongo()

    res = asyncio.run(_run())
    typer.echo(res)


def _default_service_id_from_resources_path(resources_path: str) -> str:
    # e.g. "apiframeworks_api.mongo.resources:RESOURCES" -> "apiframeworks_api"
    mod = resources_path.split(":", 1)[0]
    top = mod.split(".", 1)[0]
    return top or "service"


def cmd_setup_and_prepare(
    resources_path: str = typer.Option(
        ...,
        "--resources",
        help="Dotted path to NoSqlResource(s). e.g. 'app.db.mongo:RESOURCES'",
    ),
    mongo_url: Optional[str] = typer.Option(
        None, "--mongo-url", help="Overrides env MONGO_URL for this command."
    ),
    mongo_db: Optional[str] = typer.Option(
        None, "--mongo-db", help="Overrides env MONGO_DB for this command."
    ),
    service_id: Optional[str] = typer.Option(
        None,
        "--service-id",
        help="Stable ID for this service/app. Defaults to top-level module name.",
    ),
    allow_rebind: bool = typer.Option(
        False, "--allow-rebind", help="Permit moving to a different DB."
    ),
):
    """
    Synchronous, end-to-end helper that delegates entirely to core.setup_and_prepare().
    All env resolution, validation, and DB locking are handled in core.
    """
    _apply_mongo_env(mongo_url, mongo_db)
    resources = _normalize_resources(_load_obj(resources_path))
    sid = service_id or _default_service_id_from_resources_path(resources_path)

    res = setup_and_prepare(
        resources=resources,
        service_id=sid,
        allow_rebind=allow_rebind,
    )
    typer.echo(res)


def cmd_ping(
    mongo_url: Optional[str] = typer.Option(
        None, "--mongo-url", help="Overrides env MONGO_URL for this command."
    ),
    mongo_db: Optional[str] = typer.Option(
        None, "--mongo-db", help="Overrides env MONGO_DB for this command."
    ),
):
    """
    Simple connectivity check (db.command('ping')).
    """
    _apply_mongo_env(mongo_url, mongo_db)
    prepare_process_env(".")

    import asyncio

    from svc_infra.db.nosql.mongo.client import acquire_db  # local import to avoid side effects

    async def _run():
        await init_mongo()
        try:
            db = await acquire_db()
            res = await db.command("ping")
            return {"ok": (res or {}).get("ok") == 1}
        finally:
            await close_mongo()

    res = asyncio.run(_run())
    typer.echo(res)


def register(app: typer.Typer) -> None:
    app.command("mongo-prepare")(cmd_prepare)
    app.command("mongo-setup-and-prepare")(cmd_setup_and_prepare)
    app.command("mongo-ping")(cmd_ping)



from typing import Any, Callable, Optional

from svc_infra.db.sql.service import SqlService

PreHook = Callable[[dict[str, Any]], dict[str, Any]]


class SqlServiceWithHooks(SqlService):
    def __init__(
        self,
        repo,
        pre_create: Optional[PreHook] = None,
        pre_update: Optional[PreHook] = None,
    ):
        super().__init__(repo)
        self._pre_create = pre_create
        self._pre_update = pre_update

    async def pre_create(self, data: dict[str, Any]) -> dict[str, Any]:
        return self._pre_create(data) if self._pre_create else data

    async def pre_update(self, data: dict[str, Any]) -> dict[str, Any]:
        return self._pre_update(data) if self._pre_update else data



from __future__ import annotations

import base64
import hashlib
import os
from datetime import datetime, timezone

import pyotp
from fastapi import APIRouter, Body, Depends, HTTPException, Request, status
from fastapi_users import FastAPIUsers
from pydantic import BaseModel
from sqlalchemy import select
from starlette.responses import JSONResponse

from svc_infra.api.fastapi import public_router
from svc_infra.api.fastapi.auth.pre_auth import get_mfa_pre_jwt_writer
from svc_infra.api.fastapi.auth.sender import get_sender
from svc_infra.api.fastapi.auth.settings import get_auth_settings
from svc_infra.api.fastapi.db.sql.session import SqlSessionDep

# --- Email OTP store (replace with Redis in prod) ---
EMAIL_OTP_STORE: dict[str, dict] = {}  # key = uid (or jti), value={hash,exp,attempts,next_send}


# ---- DTOs ----
class StartSetupOut(BaseModel):
    otpauth_url: str
    secret: str
    qr_svg: str | None = None  # optional: inline SVG


class ConfirmSetupIn(BaseModel):
    code: str


class VerifyMFAIn(BaseModel):
    code: str
    pre_token: str


class DisableMFAIn(BaseModel):
    code: str | None = None
    recovery_code: str | None = None


class RecoveryCodesOut(BaseModel):
    codes: list[str]


class SendEmailCodeIn(BaseModel):
    pre_token: str


class SendEmailCodeOut(BaseModel):
    sent: bool = True
    cooldown_seconds: int = 60


class MFAStatusOut(BaseModel):
    enabled: bool
    methods: list[str]
    confirmed_at: datetime | None = None
    email_mask: str | None = None
    email_otp: dict | None = None


# ---- Utils ----
def _qr_svg_from_uri(uri: str) -> str:
    # Placeholder SVG; most frontends will render their own QR
    return (
        "<svg xmlns='http://www.w3.org/2000/svg' width='280' height='280'>"
        "<rect width='100%' height='100%' fill='#fff'/>"
        f"<text x='10' y='20' font-size='10'>{uri}</text></svg>"
    )


def _random_base32() -> str:
    return pyotp.random_base32(length=32)


def _gen_recovery_codes(n: int, length: int) -> list[str]:
    out = []
    for _ in range(n):
        raw = base64.urlsafe_b64encode(os.urandom(24)).decode().rstrip("=")
        out.append(raw[:length])
    return out


def _hash(s: str) -> str:
    return hashlib.sha256(s.encode()).hexdigest()


def _gen_numeric_code(n: int = 6) -> str:
    import random

    return "".join(str(random.randrange(10)) for _ in range(n))


def _now_utc_ts() -> int:
    from datetime import datetime, timezone

    return int(datetime.now(timezone.utc).timestamp())


# ---- Router factory ----
def mfa_router(
    *,
    user_model: type,
    get_strategy,  # from get_fastapi_users()
    fapi: FastAPIUsers,
    auth_prefix: str = "/auth",
) -> APIRouter:
    router = public_router(prefix=f"{auth_prefix}/mfa", tags=["auth:mfa"])

    # Resolve current user via cookie OR bearer, using fastapi-users v10 strategy.read_token(..., user_manager)
    async def _get_user_and_session(
        request: Request,
        session: SqlSessionDep,
        user_manager=Depends(fapi.get_user_manager),
    ):
        st = get_auth_settings()
        token = request.headers.get("authorization", "").removeprefix(
            "Bearer "
        ).strip() or request.cookies.get(st.auth_cookie_name)
        if not token:
            raise HTTPException(401, "Missing token")

        strategy = get_strategy()
        try:
            user = await strategy.read_token(token, user_manager)  # fastapi-users user
            if not user:
                raise HTTPException(401, "Invalid token")
        except Exception:
            raise HTTPException(401, "Invalid token")

        # IMPORTANT: rehydrate into *your* session
        db_user = await session.get(user_model, user.id)
        if not db_user:
            raise HTTPException(401, "Invalid token")

        return db_user, session

    @router.post(
        "/start",
        response_model=StartSetupOut,
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def start_setup(user_sess=Depends(_get_user_and_session)):
        user, session = user_sess

        if getattr(user, "mfa_enabled", False):
            raise HTTPException(400, "MFA already enabled")

        st = get_auth_settings()
        secret = _random_base32()
        issuer = st.mfa_issuer
        label = getattr(user, "email", None) or f"user-{user.id}"
        uri = pyotp.totp.TOTP(secret).provisioning_uri(name=label, issuer_name=issuer)

        # Update and COMMIT
        user.mfa_secret = secret
        user.mfa_enabled = False
        user.mfa_confirmed_at = None
        await session.commit()

        # (Optional) verify it actually persisted:
        # fresh_secret = (await session.execute(
        #     select(user_model.mfa_secret).where(user_model.id == user.id)
        # )).scalar_one()
        # assert fresh_secret == secret

        return StartSetupOut(otpauth_url=uri, secret=secret, qr_svg=_qr_svg_from_uri(uri))

    @router.post(
        "/confirm",
        response_model=RecoveryCodesOut,
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def confirm_setup(
        payload: ConfirmSetupIn = Body(...), user_sess=Depends(_get_user_and_session)
    ):
        user, session = user_sess

        # RELOAD from DB to avoid stale state
        user = (
            await session.execute(select(user_model).where(user_model.id == user.id))
        ).scalar_one()

        if not getattr(user, "mfa_secret", None):
            raise HTTPException(400, "No setup in progress")

        totp = pyotp.TOTP(user.mfa_secret)
        if not totp.verify(payload.code, valid_window=1):
            raise HTTPException(400, "Invalid code")

        st = get_auth_settings()
        codes = _gen_recovery_codes(st.mfa_recovery_codes, st.mfa_recovery_code_length)

        user.mfa_recovery = [_hash(c) for c in codes]
        user.mfa_enabled = True
        user.mfa_confirmed_at = datetime.now(timezone.utc)
        await session.commit()

        return RecoveryCodesOut(codes=codes)

    @router.post(
        "/disable",
        status_code=status.HTTP_204_NO_CONTENT,
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def disable_mfa(
        payload: DisableMFAIn = Body(...),
        user_sess=Depends(_get_user_and_session),
    ):
        user, session = user_sess
        if not getattr(user, "mfa_enabled", False):
            return JSONResponse(status_code=204, content={})

        ok = False
        if payload.code and getattr(user, "mfa_secret", None):
            totp = pyotp.TOTP(user.mfa_secret)
            ok = totp.verify(payload.code, valid_window=1)

        if not ok and payload.recovery_code and getattr(user, "mfa_recovery", None):
            dig = _hash(payload.recovery_code)
            if dig in user.mfa_recovery:
                user.mfa_recovery.remove(dig)  # burn one
                ok = True

        if not ok:
            raise HTTPException(400, "Invalid code")

        user.mfa_enabled = False
        user.mfa_secret = None
        user.mfa_recovery = None
        user.mfa_confirmed_at = None
        await session.commit()
        return JSONResponse(status_code=204, content={})

    @router.post("/verify")
    async def verify_mfa(
        session: SqlSessionDep,
        payload: VerifyMFAIn = Body(...),
    ):
        st = get_auth_settings()
        strategy = get_strategy()

        # 1) read/verify pre-auth token (aud = mfa)
        try:
            pre = await get_mfa_pre_jwt_writer().read(payload.pre_token)
            uid = pre.get("sub")
            if not uid:
                raise HTTPException(401, "Invalid pre-auth token")
        except Exception:
            raise HTTPException(401, "Invalid pre-auth token")

        # 2) load user
        user = await session.get(user_model, uid)
        if not user:
            raise HTTPException(401, "Invalid pre-auth token")

        # NEW: block disabled accounts here with a clear error
        if not getattr(user, "is_active", True):
            raise HTTPException(401, "account_disabled")

        if (not getattr(user, "mfa_enabled", False)) or (not getattr(user, "mfa_secret", None)):
            raise HTTPException(401, "MFA not enabled")

        # 3) verify TOTP or fallback
        ok = False

        # A) TOTP
        totp = pyotp.TOTP(user.mfa_secret)
        if totp.verify(payload.code, valid_window=1):
            ok = True
        else:
            # B) Recovery code
            dig = _hash(payload.code)
            if getattr(user, "mfa_recovery", None) and dig in user.mfa_recovery:
                user.mfa_recovery.remove(dig)
                await session.commit()  # persist burn
                ok = True
            else:
                # C) Email OTP (bound to uid via pre_token above)
                rec = EMAIL_OTP_STORE.get(str(uid))
                now = _now_utc_ts()
                if rec:
                    if (
                        now <= rec["exp"]
                        and rec["attempts_left"] > 0
                        and _hash(payload.code) == rec["hash"]
                    ):
                        ok = True
                        EMAIL_OTP_STORE.pop(str(uid), None)  # burn on success
                    else:
                        rec["attempts_left"] = max(0, rec["attempts_left"] - 1)

        if not ok:
            raise HTTPException(400, "Invalid code")

        # NEW: set last_login on successful MFA
        user.last_login = datetime.now(timezone.utc)
        await session.commit()

        # 4) mint normal JWT and set cookie
        token = await strategy.write_token(user)
        resp = JSONResponse({"access_token": token, "token_type": "bearer"})
        resp.set_cookie(
            key=st.auth_cookie_name,
            value=token,
            max_age=st.session_cookie_max_age_seconds,
            httponly=True,
            secure=bool(st.session_cookie_secure),
            samesite=str(st.session_cookie_samesite).lower(),
            domain=(getattr(st, "session_cookie_domain", None) or None),
            path="/",
        )
        return resp

    @router.post(
        "/send_code",
        response_model=SendEmailCodeOut,
        description="Sends a 6-digit email OTP tied to the `pre_token`. Returns a resend cooldown.",
    )
    async def send_email_code(
        session: SqlSessionDep,
        payload: SendEmailCodeIn = Body(...),
    ):
        # 1) Validate pre_token and extract uid
        try:
            pre = await get_mfa_pre_jwt_writer().read(payload.pre_token)
            uid = pre.get("sub")
            if not uid:
                raise HTTPException(401, "Invalid pre-auth token")
        except Exception:
            raise HTTPException(401, "Invalid pre-auth token")

        # 1b) Load user to get their email
        user = await session.get(user_model, uid)
        if not user or not getattr(user, "email", None):
            # (optionally also check user.mfa_enabled here)
            raise HTTPException(401, "Invalid pre-auth token")

        st = get_auth_settings()
        now = _now_utc_ts()
        ttl = getattr(st, "email_otp_ttl_seconds", 5 * 60)
        cooldown = getattr(st, "email_otp_cooldown_seconds", 60)
        max_attempts = getattr(st, "email_otp_attempts", 5)

        # 2) Throttle resends
        rec = EMAIL_OTP_STORE.get(str(uid))
        if rec and rec.get("next_send") and now < rec["next_send"]:
            return SendEmailCodeOut(sent=True, cooldown_seconds=rec["next_send"] - now)

        # 3) Generate + store (hashed) OTP
        code = _gen_numeric_code(6)
        EMAIL_OTP_STORE[str(uid)] = {
            "hash": _hash(code),
            "exp": now + ttl,
            "attempts_left": max_attempts,
            "next_send": now + cooldown,
        }

        # 4) Send email
        sender = get_sender()
        sender.send(
            to=user.email,
            subject="Your sign-in code",
            html_body=f"""
                <p>Your code is: <b>{code}</b></p>
                <p>It expires in {ttl // 60} minutes.</p>
                <p>If you didn’t request this, you can ignore this email.</p>
            """,
        )

        return SendEmailCodeOut(sent=True, cooldown_seconds=cooldown)

    @router.get(
        "/status",
        response_model=MFAStatusOut,
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def mfa_status(user_sess=Depends(_get_user_and_session)):
        user, _ = user_sess
        enabled = bool(getattr(user, "mfa_enabled", False))
        confirmed_at = getattr(user, "mfa_confirmed_at", None)

        methods = []
        if enabled and getattr(user, "mfa_secret", None):
            methods.append("totp")
            methods.append("recovery")
        # Email OTP is always offered in your flow at verify-time
        methods.append("email")

        def _mask(email: str) -> str:
            if not email or "@" not in email:
                return None
            name, domain = email.split("@", 1)
            if len(name) <= 1:
                masked = "*"
            elif len(name) == 2:
                masked = name[0] + "*"
            else:
                masked = name[0] + "*" * (len(name) - 2) + name[-1]
            return f"{masked}@{domain}"

        email = getattr(user, "email", None)
        st = get_auth_settings()
        return MFAStatusOut(
            enabled=enabled,
            methods=methods,
            confirmed_at=confirmed_at,
            email_mask=_mask(email) if email else None,
            email_otp={"cooldown_seconds": st.email_otp_cooldown_seconds},
        )

    @router.post(
        "/recovery/regenerate",
        response_model=RecoveryCodesOut,
        openapi_extra={
            "security": [{"OAuth2PasswordBearer": []}, {"cookieAuth": []}, {"APIKeyHeader": []}]
        },
    )
    async def regenerate_recovery_codes(user_sess=Depends(_get_user_and_session)):
        user, session = user_sess
        if not getattr(user, "mfa_enabled", False):
            raise HTTPException(400, "MFA not enabled")

        st = get_auth_settings()
        codes = _gen_recovery_codes(st.mfa_recovery_codes, st.mfa_recovery_code_length)
        user.mfa_recovery = [_hash(c) for c in codes]
        await session.commit()
        return RecoveryCodesOut(codes=codes)

    return router



from __future__ import annotations

import os
import warnings
from enum import StrEnum
from functools import cache
from pathlib import Path
from typing import List, NamedTuple, Optional

from dotenv import load_dotenv

from svc_infra.app.root import resolve_project_root


class Environment(StrEnum):
    LOCAL = "local"
    DEV = "dev"
    TEST = "test"
    PROD = "prod"


# Handy aliases
LOCAL_ENV = Environment.LOCAL
DEV_ENV = Environment.DEV
TEST_ENV = Environment.TEST
PROD_ENV = Environment.PROD


# Map common aliases -> canonical
SYNONYMS: dict[str, Environment] = {
    "development": DEV_ENV,
    "dev": DEV_ENV,
    "local": LOCAL_ENV,
    "uat": TEST_ENV,
    "test": TEST_ENV,
    "preview": TEST_ENV,
    "staging": TEST_ENV,
    "prod": PROD_ENV,
    "production": PROD_ENV,
}

ALL_ENVIRONMENTS = {e for e in Environment}


def _normalize(raw: str | None) -> Environment | None:
    """
    Normalize raw environment string to canonical Env enum, case-insensitively.
    """
    if not raw:
        return None
    val = raw.strip().casefold()  # case-insensitive, handles unicode edge cases
    # Check against canonical enum values
    if val in (e.value for e in Environment):
        return Environment(val)  # exact match
    # Check against synonyms
    return SYNONYMS.get(val)


@cache
def get_current_environment() -> Environment:
    """
    Resolve the current environment once, with sensible fallbacks.

    Precedence:
      1) APP_ENV
      2) RAILWAY_ENVIRONMENT_NAME
      3) "local" (default)

    Unknown values fall back to LOCAL with a one-time warning.
    """
    raw = os.getenv("APP_ENV") or os.getenv("RAILWAY_ENVIRONMENT_NAME")
    env = _normalize(raw)
    if env is None:
        if raw:
            warnings.warn(
                f"Unrecognized environment '{raw}', defaulting to 'local'.",
                RuntimeWarning,
                stacklevel=2,
            )
        env = LOCAL_ENV
    return env


class EnvironmentFlags(NamedTuple):
    environment: Environment
    is_local: bool
    is_dev: bool
    is_test: bool
    is_prod: bool


def get_environment_flags(environment: Environment | None = None) -> EnvironmentFlags:
    e = environment or get_current_environment()
    return EnvironmentFlags(
        environment=e,
        is_local=(e == _normalize("local")),
        is_dev=(e == _normalize("dev")),
        is_test=(e == _normalize("test")),
        is_prod=(e == _normalize("prod")),
    )


# Handy globals
CURRENT_ENVIRONMENT: Environment = get_current_environment()
ENV_FLAGS: EnvironmentFlags = get_environment_flags(CURRENT_ENVIRONMENT)
IS_LOCAL, IS_DEV, IS_TEST, IS_PROD = (
    ENV_FLAGS.is_local,
    ENV_FLAGS.is_dev,
    ENV_FLAGS.is_test,
    ENV_FLAGS.is_prod,
)


def pick(*, prod, nonprod=None, dev=None, test=None, local=None):
    """
    Choose a value based on the active environment.

    Example:
        log_level = pick(prod="INFO", nonprod="DEBUG", dev="DEBUG")
    """
    e = get_current_environment()
    if e is PROD_ENV:
        return prod
    if e is DEV_ENV and dev is not None:
        return dev
    if e is TEST_ENV and test is not None:
        return test
    if e is LOCAL_ENV and local is not None:
        return local
    if nonprod is not None:
        return nonprod
    raise ValueError("pick(): No value found for environment and 'nonprod' was not provided.")


def find_env_file(start: Optional[Path] = None) -> Optional[Path]:
    env_file = os.getenv("APP_ENV_FILE") or os.getenv("SVC_INFRA_ENV_FILE")
    if env_file:
        p = Path(env_file).expanduser()
        return p if p.exists() else None

    cur = (start or Path.cwd()).resolve()
    for p in [cur, *cur.parents]:
        candidate = p / ".env"
        if candidate.exists():
            return candidate
    return None


def load_env_if_present(path: Optional[Path], *, override: bool = False) -> List[str]:
    if not path:
        return []
    before = dict(os.environ)
    load_dotenv(dotenv_path=path, override=override)
    changed = []
    for k, v in os.environ.items():
        if k not in before or before.get(k) != v:
            changed.append(k)
    return sorted(changed)


def prepare_env() -> Path:
    """
    Return (project_root, debug_note). No chdir here; runner handles cwd.
    """
    root = resolve_project_root()
    env_file = find_env_file(start=root)
    load_env_if_present(env_file, override=False)
    return root



import logging
import traceback

from fastapi import Request
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.responses import JSONResponse
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from starlette.exceptions import HTTPException as StarletteHTTPException

from svc_infra.api.fastapi.middleware.errors.exceptions import FastApiException
from svc_infra.app.env import IS_PROD

logger = logging.getLogger(__name__)


def get_error_type(exc: Exception) -> str:
    return getattr(exc, "error", None) or type(exc).__name__


def log_exception(level: int, exc: Exception, request: Request, status_code: int) -> None:
    logger.log(
        level,
        f"{request.method} {request.url.path} [{status_code}] {type(exc).__name__}: {exc}",
        exc_info=True,
    )


def format_error_response(
    exc: Exception, request: Request, status_code: int, detail: str | dict | list
) -> JSONResponse:
    error_type = get_error_type(exc)
    log_exception(logging.ERROR, exc, request, status_code)

    response_content = {
        "error": error_type,
        "detail": (
            detail
            if not IS_PROD
            else ("Something went wrong. Please contact support." if status_code == 500 else detail)
        ),
    }

    if not IS_PROD and status_code == 500:
        response_content["trace"] = traceback.format_exc()

    return JSONResponse(status_code=status_code, content=response_content)


def register_error_handlers(app):
    @app.exception_handler(FastApiException)
    async def handle_api_frameworks_exception(request: Request, exc: FastApiException):
        return format_error_response(exc, request, exc.status_code, exc.detail)

    @app.exception_handler(RequestValidationError)
    async def handle_validation_error(request: Request, exc: RequestValidationError):
        detail = exc.errors() if not IS_PROD else "Invalid request payload."
        return format_error_response(exc, request, 422, detail)

    @app.exception_handler(HTTPException)
    async def handle_http_exception(request: Request, exc: HTTPException):
        return format_error_response(exc, request, exc.status_code, exc.detail)

    @app.exception_handler(StarletteHTTPException)
    async def handle_starlette_http_exception(request: Request, exc: StarletteHTTPException):
        return format_error_response(exc, request, exc.status_code, exc.detail)

    @app.exception_handler(SQLAlchemyError)
    async def handle_sqlalchemy_error(request: Request, exc: SQLAlchemyError):
        return format_error_response(exc, request, 500, "Please try again later.")

    @app.exception_handler(Exception)
    async def handle_unexpected_error(request: Request, exc: Exception):
        return format_error_response(exc, request, 500, str(exc))

    @app.exception_handler(IntegrityError)
    async def handle_integrity_error(_: Request, exc: IntegrityError):
        msg = str(getattr(exc, "orig", exc))
        if "duplicate key value" in msg or "UniqueViolation" in msg:
            return JSONResponse(status_code=409, content={"detail": "Record already exists."})
        if "not-null" in msg or "NotNullViolation" in msg:
            return JSONResponse(status_code=400, content={"detail": "Missing required field."})
        return JSONResponse(status_code=500, content={"detail": "Database error."})



from __future__ import annotations

from pydantic import BaseModel


class DocumentBase(BaseModel):
    """Marker base for Pydantic document models used for auto schema derivation."""

    model_config = {"from_attributes": True}



from __future__ import annotations

import os
import socket
import subprocess
from pathlib import Path
from urllib.parse import urlparse

import typer

# --- NEW: load .env automatically (best-effort) ---
try:
    from dotenv import load_dotenv  # type: ignore
except Exception:  # pragma: no cover
    load_dotenv = None

from svc_infra.obs.cloud_dash import push_dashboards_from_pkg
from svc_infra.utils import render_template, write


def _run(cmd: list[str], *, env: dict | None = None):
    subprocess.run(cmd, check=True, env=env)


def _emit_local_stack(root: Path, metrics_url: str):
    write(
        root / "docker-compose.yml",
        render_template("svc_infra.obs.providers.grafana.templates", "docker-compose.yml.tmpl", {}),
    )
    p = urlparse(metrics_url)
    prom_yml = render_template(
        "svc_infra.obs.providers.grafana.templates",
        "prometheus.yml.tmpl",
        {
            "metrics_path": (p.path or "/metrics"),
            "target": (p.netloc or "host.docker.internal:8000"),
        },
    )
    write(root / "prometheus.yml", prom_yml)

    # provisioning + dashboards
    root.joinpath("provisioning/datasources").mkdir(parents=True, exist_ok=True)
    root.joinpath("provisioning/dashboards").mkdir(parents=True, exist_ok=True)
    root.joinpath("dashboards").mkdir(parents=True, exist_ok=True)

    from importlib.resources import files

    tpl = files("svc_infra.obs.providers.grafana")
    write(
        root / "provisioning/datasources/datasource.yml",
        tpl.joinpath("templates/provisioning/datasource.yml").read_text(),
    )
    write(
        root / "provisioning/dashboards/dashboards.yml",
        tpl.joinpath("templates/provisioning/dashboards.yml").read_text(),
    )
    for d in tpl.joinpath("dashboards").iterdir():
        if d.name.endswith(".json"):
            write(root / "dashboards" / d.name, d.read_text())


def _emit_local_agent(root: Path, metrics_url: str):
    """
    Render the agent + compose from the dedicated compose_cloud templates.
    """
    # Write agent.yaml from template (no hardcoded strings)
    agent_text = render_template(
        tmpl_dir="svc_infra.obs.providers.compose_cloud.templates",
        name="agent.yaml.tmpl",
        subs={},  # values come from env at runtime via ${...}
    )
    write(root / "agent.yaml", agent_text, overwrite=True)

    compose_text = render_template(
        tmpl_dir="svc_infra.obs.providers.compose_cloud.templates",
        name="docker-compose.cloud.yml.tmpl",
        subs={},  # all env-driven
    )
    write(root / "docker-compose.cloud.yml", compose_text, overwrite=True)


def _port_free(p: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.connect_ex(("127.0.0.1", p)) != 0


def _choose_port(preferred: int, limit: int = 15) -> int:
    p = preferred
    for _ in range(limit):
        if _port_free(p):
            return p
        p += 1
    return preferred


def up():
    """
    Auto-detect mode:
      - If GRAFANA_CLOUD_URL & GRAFANA_CLOUD_TOKEN → Cloud mode (push dashboards).
        If remote_write creds present → also run local Agent to push metrics.
      - Else → Local mode (Grafana + Prometheus).
    """
    # NEW: load .env once, best-effort, without crashing if package missing
    if load_dotenv:
        try:
            load_dotenv(dotenv_path=Path(".env"), override=False)
        except Exception:
            pass

    root = Path(".obs")
    root.mkdir(exist_ok=True)
    metrics_url = os.getenv("SVC_INFRA_METRICS_URL", "http://host.docker.internal:8000/metrics")

    cloud_url = os.getenv("GRAFANA_CLOUD_URL", "").strip()
    cloud_token = os.getenv("GRAFANA_CLOUD_TOKEN", "").strip()

    if cloud_url and cloud_token:
        folder = os.getenv("SVC_INFRA_CLOUD_FOLDER", "Service Infrastructure")
        push_dashboards_from_pkg(cloud_url, cloud_token, folder)
        typer.echo(f"[cloud] dashboards synced to '{folder}'")

        # NOTE: look for RW token (not the Grafana API token)
        if all(
            os.getenv(k)
            for k in (
                "GRAFANA_CLOUD_PROM_URL",
                "GRAFANA_CLOUD_PROM_USERNAME",
                "GRAFANA_CLOUD_RW_TOKEN",
            )
        ):
            _emit_local_agent(root, metrics_url)
            _run(
                ["docker", "compose", "-f", str(root / "docker-compose.cloud.yml"), "up", "-d"],
                env=os.environ.copy(),
            )
            typer.echo("[cloud] local Grafana Agent started (pushing metrics to Cloud)")
        else:
            typer.echo("[cloud] expecting Agent sidecar in deployment to push metrics")
        return

    # Local mode
    local_graf = _choose_port(int(os.getenv("GRAFANA_PORT", "3000")))
    local_prom = _choose_port(int(os.getenv("PROM_PORT", "9090")))
    env = os.environ.copy()
    env["GRAFANA_PORT"] = str(local_graf)
    env["PROM_PORT"] = str(local_prom)
    _emit_local_stack(root, metrics_url)
    _run(["docker", "compose", "-f", str(root / "docker-compose.yml"), "up", "-d"], env=env)
    typer.echo(f"Local Grafana → http://localhost:{local_graf}  (admin/admin)")
    typer.echo(f"Local Prometheus → http://localhost:{local_prom}")


def down():
    root = Path(".obs")
    if (root / "docker-compose.yml").exists():
        subprocess.run(
            ["docker", "compose", "-f", str(root / "docker-compose.yml"), "down"], check=False
        )
    if (root / "docker-compose.cloud.yml").exists():
        subprocess.run(
            ["docker", "compose", "-f", str(root / "docker-compose.cloud.yml"), "down"], check=False
        )
    typer.echo("Stopped local obs services.")


def scaffold(target: str = typer.Option(..., help="compose|railway|k8s|fly")):
    from importlib.resources import files

    out = Path("obs-sidecar") / target
    out.mkdir(parents=True, exist_ok=True)

    base = files("svc_infra.obs.templates.sidecars").joinpath(target)
    for p in base.rglob("*"):
        if p.is_file():
            rel = p.relative_to(base)
            dst = out / rel
            dst.parent.mkdir(parents=True, exist_ok=True)
            dst.write_text(p.read_text(encoding="utf-8"), encoding="utf-8")

    typer.echo(f"Wrote sidecar template to {out}. Fill envs and deploy.")


def register(app: typer.Typer) -> None:
    app.command("obs-up")(up)
    app.command("obs-down")(down)
    app.command("obs-scaffold")(scaffold)



from __future__ import annotations

from typing import Any, AsyncIterator, Callable, Tuple
from uuid import UUID

from fastapi_users import FastAPIUsers
from fastapi_users.authentication import AuthenticationBackend, BearerTransport, JWTStrategy
from fastapi_users.manager import BaseUserManager, UUIDIDMixin

from svc_infra.api.fastapi import DualAPIRouter, dualize_router
from svc_infra.api.fastapi.auth.settings import get_auth_settings
from svc_infra.api.fastapi.deps import Require
from svc_infra.app.env import CURRENT_ENVIRONMENT, DEV_ENV, LOCAL_ENV

from ...auth.sender import get_sender
from .session import SqlSessionDep


def get_fastapi_users(
    user_model: Any,
    user_schema_read: Any,
    user_schema_create: Any,
    user_schema_update: Any,
    *,
    public_auth_prefix: str = "/auth",
) -> Tuple[
    FastAPIUsers,
    AuthenticationBackend,
    DualAPIRouter,
    DualAPIRouter,
    Callable,
    DualAPIRouter,
    DualAPIRouter,
    DualAPIRouter,
]:
    from fastapi_users_db_sqlalchemy import SQLAlchemyUserDatabase

    async def get_user_db(session: SqlSessionDep) -> AsyncIterator[Any]:
        yield SQLAlchemyUserDatabase(session, user_model)

    class UserManager(UUIDIDMixin, BaseUserManager[Any, UUID]):
        reset_password_token_secret = "unused"
        verification_token_secret = "unused"

        async def on_after_register(self, user: Any, request=None):
            st = get_auth_settings()
            if CURRENT_ENVIRONMENT in (DEV_ENV, LOCAL_ENV) and bool(st.auto_verify_in_dev):
                await self.user_db.update(user, {"is_verified": True})
                return
            await self.request_verify(user, request)

        async def on_after_request_verify(self, user: Any, token: str, request=None):
            verify_url = f"{public_auth_prefix}/verify?token={token}"
            sender = get_sender()
            sender.send(
                to=getattr(user, "email"),
                subject="Verify your account",
                html_body=f"""
                    <p>Hi {getattr(user, 'full_name', '') or 'there'},</p>
                    <p>Click to verify your account:</p>
                    <p><a href="{verify_url}">{verify_url}</a></p>
                """,
            )

        async def on_after_forgot_password(self, user: Any, token: str, request=None):
            reset_url = f"{public_auth_prefix}/reset-password?token={token}"
            sender = get_sender()
            sender.send(
                to=getattr(user, "email"),
                subject="Reset your password",
                html_body=f"""
                    <p>We received a request to reset your password.</p>
                    <p><a href="{reset_url}">{reset_url}</a></p>
                    <p>If you didn’t request this, you can ignore this email.</p>
                """,
            )

        async def on_after_reset_password(self, user: Any, request=None):
            # Optional: audit/log, notify, etc. Keep it no-op by default.
            pass

    async def get_user_manager(user_db=Require(get_user_db)):
        yield UserManager(user_db)

    def get_jwt_strategy() -> JWTStrategy:
        st = get_auth_settings()
        jwt_block = getattr(st, "jwt", None)
        if jwt_block and getattr(jwt_block, "secret", None):
            secret = jwt_block.secret.get_secret_value()
        else:
            secret = "svc-dev-secret-change-me"
        lifetime = getattr(jwt_block, "lifetime_seconds", None) if jwt_block else None
        if not isinstance(lifetime, int) or lifetime <= 0:
            lifetime = 3600
        return JWTStrategy(secret=secret, lifetime_seconds=lifetime)

    bearer_transport = BearerTransport(tokenUrl=f"{public_auth_prefix}/login")
    auth_backend = AuthenticationBackend(
        name="jwt", transport=bearer_transport, get_strategy=get_jwt_strategy
    )
    fastapi_users = FastAPIUsers(get_user_manager, [auth_backend])

    # IMPORTANT: requires_verification=True forces unverified users to be rejected on login.
    auth_router = dualize_router(
        fastapi_users.get_auth_router(auth_backend, requires_verification=True)
    )
    users_router = dualize_router(
        fastapi_users.get_users_router(user_schema_read, user_schema_create, user_schema_update)
    )
    register_router = dualize_router(
        fastapi_users.get_register_router(user_schema_read, user_schema_create)
    )
    verify_router = dualize_router(fastapi_users.get_verify_router(user_schema_read))
    reset_router = dualize_router(fastapi_users.get_reset_password_router())

    return (
        fastapi_users,
        auth_backend,
        auth_router,
        users_router,
        get_jwt_strategy,
        register_router,
        verify_router,
        reset_router,
    )



from .env import pick
from .logging import setup_logging
from .logging.formats import LoggingConfig, LogLevelOptions

__all__ = [
    "setup_logging",
    "LoggingConfig",
    "LogLevelOptions",
    "pick",
]



from typing import Annotated, Any, Optional, Sequence, Type, TypeVar, cast

from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import BaseModel

from svc_infra.api.fastapi import public_router
from svc_infra.api.fastapi.db.http import (
    LimitOffsetParams,
    OrderParams,
    Page,
    SearchParams,
    build_order_by,
    dep_limit_offset,
    dep_order,
    dep_search,
)
from svc_infra.db.sql.service import SqlService

from .session import SqlSessionDep

CreateModel = TypeVar("CreateModel", bound=BaseModel)
ReadModel = TypeVar("ReadModel", bound=BaseModel)
UpdateModel = TypeVar("UpdateModel", bound=BaseModel)


def make_crud_router_plus_sql(
    *,
    model: type[Any],
    service: SqlService,
    read_schema: Type[ReadModel],
    create_schema: Type[CreateModel],
    update_schema: Type[UpdateModel],
    prefix: str,
    tags: list[str] | None = None,
    search_fields: Optional[Sequence[str]] = None,
    default_ordering: Optional[str] = None,
    allowed_order_fields: Optional[list[str]] = None,
    mount_under_db_prefix: bool = True,
) -> APIRouter:
    router_prefix = ("/_sql" + prefix) if mount_under_db_prefix else prefix
    router = public_router(
        prefix=router_prefix,
        tags=tags or [prefix.strip("/")],
        redirect_slashes=False,
    )

    def _parse_ordering_to_fields(order_spec: Optional[str]) -> list[str]:
        if not order_spec:
            return []
        pieces = [p.strip() for p in order_spec.split(",") if p.strip()]
        fields: list[str] = []
        for p in pieces:
            name = p[1:] if p.startswith("-") else p
            if allowed_order_fields and name not in (allowed_order_fields or []):
                continue
            fields.append(p)
        return fields

    # -------- LIST --------
    @router.get("", response_model=cast(Any, Page[read_schema]))
    async def list_items(
        lp: Annotated[LimitOffsetParams, Depends(dep_limit_offset)],
        op: Annotated[OrderParams, Depends(dep_order)],
        sp: Annotated[SearchParams, Depends(dep_search)],
        session: SqlSessionDep,  # type: ignore[name-defined]
    ):
        order_spec = op.order_by or default_ordering
        order_fields = _parse_ordering_to_fields(order_spec)
        order_by = build_order_by(model, order_fields)

        if sp.q:
            fields = [
                f.strip()
                for f in (sp.fields or (",".join(search_fields or []) or "")).split(",")
                if f.strip()
            ]
            items = await service.search(
                session, q=sp.q, fields=fields, limit=lp.limit, offset=lp.offset, order_by=order_by
            )
            total = await service.count_filtered(session, q=sp.q, fields=fields)
        else:
            items = await service.list(session, limit=lp.limit, offset=lp.offset, order_by=order_by)
            total = await service.count(session)
        return Page[read_schema].from_items(
            total=total, items=items, limit=lp.limit, offset=lp.offset
        )

    # -------- GET by id --------
    @router.get("/{item_id}", response_model=cast(Any, read_schema))
    async def get_item(item_id: Any, session: SqlSessionDep):  # type: ignore[name-defined]
        row = await service.get(session, item_id)
        if not row:
            raise HTTPException(404, "Not found")
        return row

    # -------- CREATE --------
    @router.post("", response_model=cast(Any, read_schema), status_code=201)
    async def create_item(
        session: SqlSessionDep,  # type: ignore[name-defined]
        payload: create_schema = Body(...),
    ):
        data = cast(BaseModel, payload).model_dump(exclude_unset=True)
        return await service.create(session, data)

    # -------- UPDATE --------
    @router.patch("/{item_id}", response_model=cast(Any, read_schema))
    async def update_item(
        item_id: Any,
        session: SqlSessionDep,  # type: ignore[name-defined]
        payload: update_schema = Body(...),
    ):
        data = cast(BaseModel, payload).model_dump(exclude_unset=True)
        row = await service.update(session, item_id, data)
        if not row:
            raise HTTPException(404, "Not found")
        return row

    # -------- DELETE --------
    @router.delete("/{item_id}", status_code=204)
    async def delete_item(item_id: Any, session: SqlSessionDep):  # type: ignore[name-defined]
        ok = await service.delete(session, item_id)
        if not ok:
            raise HTTPException(404, "Not found")
        return

    return router


__all__ = ["make_crud_router_plus_sql"]
